import os.path

EOF_MARKER = "$$$EOF$$$"
KEEPALIVE_MARKER = "$$$KEEPALIVE$$$"
CLIENT_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".cstar_perf", "client.conf")
SERVER_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".cstar_perf", "server.conf")
SERVER_KEY_PATH = os.path.join(os.path.expanduser("~"),'.cstar_perf','server.conf')



import unittest
import uuid
import string
import random
import time

from ..model import Model, Cluster, NoTestsScheduledError, APIKeyExistsError, UnknownAPIKeyError

## This isn't a true unit test, but an integration test against a real
## C* instance. It uses a separate keyspace so it shouldn't interfere
## with any existing data.

class TestModel(unittest.TestCase):
    def setUp(self):
        ks = "test_model_non_prod"
        cluster=Cluster(['127.0.0.1'])
        sess = cluster.connect()
        sess.execute("DROP KEYSPACE IF EXISTS {ks}".format(ks=ks))
        self.model = Model(cluster=cluster, keyspace=ks)

    def tearDown(self):
        sess = self.model.get_session()
        sess.execute("DROP KEYSPACE IF EXISTS {ks}".format(ks=self.model.keyspace))

    def test_instantiation(self):
        pass

    def test_status(self):
        m = self.model
        def mock_test_definition(title, cluster='bdplab', user='ryan'):
            return {'title': title,
                    'cluster': cluster,
                    'user': user}

        test1 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('1'))
        test2 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('2'))
        test3 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('3'))
        test4 = m.schedule_test(uuid.uuid1(), 'ryan', 'austin', mock_test_definition('1b'))
        test5 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('4'))
        test6 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('5'))
        test7 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('6'))
        test8 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('7'))
        test9 = m.schedule_test(uuid.uuid1(), 'ryan', 'austin', mock_test_definition('2b'))
        test10 = m.schedule_test(uuid.uuid1(), 'ryan', 'austin', mock_test_definition('3b'))
        test11 = m.schedule_test(uuid.uuid1(), 'ryan', 'austin', mock_test_definition('4b'))
        test12 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('8'))

        m.update_test_status(test1, 'in_progress')
        m.update_test_status(test2, 'in_progress')
        m.update_test_status(test4, 'in_progress')

        self.assertEquals([r['title'] for r in m.get_scheduled_tests('bdplab')], ['3','4','5','6','7','8'])
        self.assertEquals([r['title'] for r in m.get_scheduled_tests('bdplab',limit=2)], ['3','4'])
        self.assertEquals([r['title'] for r in m.get_scheduled_tests('austin')], ['2b','3b','4b'])
        self.assertEquals([r['title'] for r in m.get_in_progress_tests('bdplab')], ['1','2'])
        self.assertEquals([r['title'] for r in m.get_in_progress_tests('austin')], ['1b'])

        self.assertEquals(m.get_next_scheduled_test('bdplab')['title'], '3')
        self.assertEquals(m.get_next_scheduled_test('austin')['title'], '2b')
        self.assertRaises(NoTestsScheduledError, m.get_next_scheduled_test, 'no_such_cluster')

        m.update_test_status(test1, 'completed')
        m.update_test_status(test4, 'completed')
        self.assertEquals([r['title'] for r in m.get_completed_tests()], ['1b','1'])
        self.assertIsNotNone(m.get_completed_tests()[0]['completed_date'])


    def test_user_tests(self):
        m = self.model
        def mock_test_definition(title, cluster='bdplab', user='ryan'):
            return {'title': title,
                    'cluster': cluster,
                    'user': user}

        test1 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('1'))
        test2 = m.schedule_test(uuid.uuid1(), 'ryan', 'bdplab', mock_test_definition('2'))
        test3 = m.schedule_test(uuid.uuid1(), 'bob', 'bdplab', mock_test_definition('3'))
        test4 = m.schedule_test(uuid.uuid1(), 'mary', 'bdplab', mock_test_definition('4'))
        test5 = m.schedule_test(uuid.uuid1(), 'mary', 'bdplab', mock_test_definition('5'))

        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('ryan')], ['2','1'])
        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('bob')], ['3'])
        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('mary')], ['5','4'])
        
        m.update_test_status(test1, 'in_progress')
        m.update_test_status(test4, 'in_progress')

        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('ryan')], ['2'])
        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('bob')], ['3'])
        self.assertEquals([r['title'] for r in m.get_user_scheduled_tests('mary')], ['5'])

        self.assertEquals([r['title'] for r in m.get_user_in_progress_tests('ryan')], ['1'])
        self.assertEquals([r['title'] for r in m.get_user_in_progress_tests('mary')], ['4'])

        m.update_test_status(test2, 'completed')
        m.update_test_status(test5, 'completed')

        self.assertEquals([r['title'] for r in m.get_user_completed_tests('ryan')], ['2'])
        self.assertEquals([r['title'] for r in m.get_user_completed_tests('mary')], ['5'])
        

    def test_users(self):
        m = self.model
        m.create_user('jack@repairman.org', 'Repairman Jack', ['user','admin','repairman'])
        jack = m.get_user('jack@repairman.org')
        self.assertEquals(jack.user_id, 'jack@repairman.org')
        self.assertEquals(jack.full_name, 'Repairman Jack')
        self.assertEquals(jack.roles, set(['user','admin','repairman']))
        
        jack_roles = m.get_user_roles('jack@repairman.org')
        self.assertEquals(jack_roles, set(['user','admin','repairman']))
    
    def test_artifacts(self):
        test_id = uuid.uuid1()
        m = self.model
        a1 = m.update_test_artifact(test_id, 'logs', 'LOG data 1', 'logs.tar.gz')
        a2 = m.update_test_artifact(test_id, 'logs-2', 'LOG data 2', 'more logs')
        a3 = m.update_test_artifact(test_id, 'graph', 'GRAPH data 1', 'stats.json')

        artifact_meta = m.get_test_artifact(test_id, 'logs')
        self.assertEqual(artifact_meta['artifact_type'], 'logs')
        artifact_meta = m.get_test_artifact(test_id, 'graph')
        self.assertEqual(artifact_meta['artifact_type'], 'graph')

        artifact_meta = m.get_test_artifacts(test_id)
        # artifacts are in alpha order by type:
        self.assertEqual(artifact_meta[0]['artifact_type'], 'graph')
        self.assertEqual(artifact_meta[1]['artifact_type'], 'logs')
        self.assertEqual(artifact_meta[2]['artifact_type'], 'logs-2')

        artifact = m.get_test_artifact_data(test_id, 'logs')
        self.assertEqual(artifact.artifact, 'LOG data 1')
        self.assertEqual(artifact.description, 'logs.tar.gz')
        artifact = m.get_test_artifact_data(test_id, 'logs-2')
        self.assertEqual(artifact.artifact, 'LOG data 2')
        self.assertEqual(artifact.description, 'more logs')
        artifact = m.get_test_artifact_data(test_id, 'graph')
        self.assertEqual(artifact.artifact, 'GRAPH data 1')
        self.assertEqual(artifact.description, 'stats.json')

    def test_clusters(self):
        m = self.model

        m.add_pub_key('new_cluster1', 'cluster', 'base64 encoded pubkey 1')
        m.add_pub_key('new_cluster2', 'cluster', 'base64 encoded pubkey 2')
        self.assertRaises(APIKeyExistsError, m.add_pub_key, 'new_cluster1', 'cluster', 'base64 encoded pubkey 1')
        
        self.assertEqual(m.get_pub_key('new_cluster1')['pubkey'], 'base64 encoded pubkey 1')
        self.assertRaises(UnknownAPIKeyError, m.get_pub_key, 'new_cluster_never_seen')




"""Generic flask utilities"""
import ConfigParser
from flask import session, request, abort, make_response
from cstar_perf.frontend.lib.util import random_token
from cstar_perf.frontend import SERVER_CONFIG_PATH

def create_app_config(config_path=SERVER_CONFIG_PATH):
    config = ConfigParser.RawConfigParser()
    config.read(config_path)
    # Ensure app has secret key for signing cookies:
    if not config.has_section('server'):
        config.add_section('server')
    if not config.has_option('server','app_secret'):
        config.set('server','app_secret', random_token())
    # Ensure app has server url:
    if not config.has_option('server', 'url'):
        config.set('server', 'url', 'http://localhost:8000')
    with open(config_path, 'w') as f:
        config.write(f)
    return config

def load_app_config(config_path=SERVER_CONFIG_PATH):
    config = ConfigParser.RawConfigParser()
    config.read(config_path)
    # Ensure app has secret key for signing cookies:
    if not config.has_section('server') or not config.has_option('server','app_secret'):
        raise AssertionError('config file ({path}) does not contain app secret'.format(SERVER_CONFIG_PATH))
    return config


def csrf_protect_app(app):
    """CSRF protection for flask apps

    Client gets token by reading meta tag set in base jinja template via {{csrf_token()}}
    All POST requests require a CSRF token in an HTTP header called X-csrf
    Tokens are valid for as long as the session exists."""

    @app.before_request
    def csrf_protect():
        if request.method == "POST":
            token = session.get('_csrf_token', None)
            header = request.headers.get('X-csrf', None)
            if not token or not header or token != header:
                abort(make_response("Invalid token", 403))

    def generate_csrf_token():
        if '_csrf_token' not in session:
            session['_csrf_token'] = random_token()
        return session['_csrf_token']

    app.jinja_env.globals['csrf_token'] = generate_csrf_token 




import os, os.path
import signal
import sys
import shlex
import subprocess
import argparse

from cstar_perf.frontend.server.util import create_app_config
from cstar_perf.frontend.lib.crypto import generate_server_keys

def run_server():
    # Initialize database before gunicorn workers startup:
    from cstar_perf.frontend.server.model import Model
    db = Model()
    del db
    
    app_path = os.path.realpath(os.path.join(os.path.dirname(
        os.path.realpath(__file__)), os.path.pardir, "server"))
    os.chdir(app_path)
    
    proc = subprocess.Popen(shlex.split("gunicorn -k flask_sockets.worker --log-file=- --workers=10 app:app"))

    # Capture SIGTERM events to shutdown child gunicorn processes..
    def on_terminate(sig, frame):
        print("Killing child processes...")
        proc.terminate()
    signal.signal(signal.SIGTERM, on_terminate)

    proc.communicate()


def main():
    parser = argparse.ArgumentParser(description='cstar_perf_server')
    parser.add_argument('--get-credentials', dest='gen_credentials',
                        action='store_true', help='Get and/or create ECDSA key for signing requests.')

    args = parser.parse_args()

    if args.gen_credentials:
        generate_server_keys()
        return

    create_app_config()
    run_server()



"""Main Application Controllers"""
import functools
import httplib2
import os.path
import uuid
import time
from functools import partial

import zmq
import json
from flask import ( Flask, render_template, request, redirect, abort, Response,
                    jsonify, make_response, session)
from apiclient.discovery import build
from oauth2client.client import ( AccessTokenRefreshError,
                                  AccessTokenCredentials,
                                  flow_from_clientsecrets,
                                  FlowExchangeError)

from app import app, db, sockets
from model import Model, UnknownUserError, UnknownTestError
from notifications import console_subscribe

import logging
log = logging.getLogger('cstar_perf.controllers')


### Google+ API:
gplus = build('plus', 'v1')
google_client_secrets = os.path.join(os.path.expanduser("~"),'.cstar_perf','client_secrets.json')
with open(google_client_secrets) as f:
    google_client_id = json.load(f)['web']['client_id']

################################################################################
#### Template functions:
################################################################################
def get_user_id():
    return session.get('user_id', None)
app.jinja_env.globals['get_user_id'] = get_user_id

def user_is_authenticated():
    return session.get('logged_in',False)
app.jinja_env.globals['user_is_authenticated'] = user_is_authenticated


################################################################################
#### Helpers
################################################################################
def user_in_role(role, user=None):
    """Find if a user is in the given role"""
    if user is None:
        user = get_user_id()
    try:
        user_roles = db.get_user_roles(get_user_id())
        if role in user_roles:
            return True
    except UnknownUserError:
        pass
    return False

def requires_auth(role):
    """Ensures the current user has the appropriate authorization before
    running the wrapped function"""
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kw):
            # Do the check:
            if user_is_authenticated():
                if user_in_role(role):
                    return function(*args, **kw)
            return make_response(render_template('access_denied.jinja2.html'), 401)
        return wrapper
    return decorator

@app.context_processor
def inject_template_variables():
    """Common variables available to all templates"""
    return dict(clusters = db.get_cluster_names(),
                google_client_id=google_client_id)

################################################################################
#### Page Controllers
################################################################################

@app.route('/')
def index():
    return render_template('index.jinja2.html')

@app.route('/login', methods=['POST'])
def login():
    """Login via Google+"""
    code = request.data

    try:
        # Upgrade the authorization code into a credentials object
        oauth_flow = flow_from_clientsecrets(google_client_secrets, scope='')
        oauth_flow.redirect_uri = 'postmessage'
        credentials = oauth_flow.step2_exchange(code)
    except FlowExchangeError:
        return make_response(
            jsonify({'error':'Failed to upgrade the authorization code.'}), 401)

    # An ID Token is a cryptographically-signed JSON object encoded in base 64.
    # Normally, it is critical that you validate an ID Token before you use it,
    # but since you are communicating directly with Google over an
    # intermediary-free HTTPS channel and using your Client Secret to
    # authenticate yourself to Google, you can be confident that the token you
    # receive really comes from Google and is valid. If your server passes the
    # ID Token to other components of your app, it is extremely important that
    # the other components validate the token before using it.
    gplus_id = credentials.id_token['sub']
    
    stored_credentials = AccessTokenCredentials(session.get('credentials'), 
                                                request.user_agent)
    stored_gplus_id = session.get('gplus_id')
    if stored_credentials is not None and gplus_id == stored_gplus_id:
        return make_response(jsonify(
            {'success':'Current user is already connected.'}), 200)
    # Get the user's email address:
    http = httplib2.Http()
    http = credentials.authorize(http)
    # Get a list of people that this user has shared with this app.
    google_request = gplus.people().get(userId='me')
    user_obj = google_request.execute(http=http)
    email = None
    # Find the google account email:
    for e in user_obj['emails']:
        if e['type'] == 'account':
            email = e['value']
            break
    else:
        return make_response(
            jsonify({'error':'Authorization from Google failed.'}), 401)
    
    # Store the access token in the session for later use.
    session['credentials'] = credentials.access_token
    session['gplus_id'] = gplus_id
    session['logged_in'] = True
    session['user_id'] = email
    return make_response(jsonify({'success':'Successfully connected user.'}), 
                         200)

@app.route('/logout', methods=['GET','POST'])
def logout():
    for i in ['credentials','gplus_id','logged_in'] :
        try:
            session.pop(i)
        except KeyError:
            pass
    if request.method == "POST":
        return make_response(jsonify({'success':'Logged out.'}), 200)
    else:
        return redirect("/")


@app.route('/tests')
def tests():
    clusters = db.get_cluster_names()
    cluster_scheduled_tests = {}
    cluster_in_progress_tests = {}
    for c in clusters:
        scheduled_tests = db.get_scheduled_tests(c)
        if len(scheduled_tests) > 0:
            cluster_scheduled_tests[c] = scheduled_tests
        in_progress_tests = db.get_in_progress_tests(c)
        if len(in_progress_tests) > 0:
            cluster_in_progress_tests[c] = in_progress_tests
    completed_tests = db.get_completed_tests()
    return render_template('tests.jinja2.html', clusters=clusters, 
                           cluster_scheduled_tests=cluster_scheduled_tests, 
                           cluster_in_progress_tests=cluster_in_progress_tests,
                           completed_tests=completed_tests)

@app.route('/tests/user')
@requires_auth('user')
def my_tests():
    queued_tests = db.get_user_scheduled_tests(get_user_id())
    in_progress_tests = db.get_user_in_progress_tests(get_user_id())
    completed_tests = db.get_user_completed_tests(get_user_id())
    failed_tests = db.get_user_failed_tests(get_user_id(), 10)
    return render_template('user.jinja2.html', queued_tests=queued_tests, 
                           in_progress_tests=in_progress_tests, 
                           completed_tests=completed_tests,
                           failed_tests=failed_tests)

@app.route('/tests/id/<test_id>')
def view_test(test_id):
    try:
        test = db.get_test(test_id)
    except UnknownTestError:
        return make_response('Unknown Test {test_id}.'.format(test_id=test_id), 404)
    artifacts = db.get_test_artifacts(test_id)

    has_chart = False
    for a in artifacts:
        if a['artifact_type'] in ['failure','link']:
            # Proactively fetch non-blob artifacts:
            a['artifact'] = db.get_test_artifact_data(test_id, a['artifact_type']).artifact
        if a['artifact_type'] == 'stats':
            has_chart = True

    return render_template('view_test.jinja2.html', test=test, artifacts=artifacts, has_chart=has_chart)

@app.route('/tests/artifacts/<test_id>/<artifact_type>')
def get_artifact(test_id, artifact_type):
    if artifact_type == 'graph':
        return redirect("/graph?stats={test_id}".format(test_id=test_id))
    artifact, description = db.get_test_artifact_data(test_id, artifact_type)
    if description.endswith(".tar.gz"):
        mimetype = 'application/gzip'
    elif description.endswith(".json"):
        mimetype = 'application/json'
    else:
        mimetype = 'text/plain'

    return Response(response=artifact,
                    status=200,
                    mimetype=mimetype,
                    headers={"Content-Disposition": "filename={name}".format(name=description)}
    )

@app.route('/graph')
def graph():
    return render_template('graph.jinja2.html')


@app.route('/schedule', methods=['GET', 'POST'])
@requires_auth('user')
def schedule():
    """Page to schedule a test - GET to view - POST test json to pre-populate page"""
    job = None
    if request.method == "POST":
        job = request.get_json()
    return render_template('schedule.jinja2.html', job=job)

@app.route('/cluster/<cluster_name>')
@requires_auth('user')
def cluster(cluster_name):
    return render_template('cluster.jinja2.html',
                           cluster_name=cluster_name)
    
################################################################################
#### JSON API
################################################################################

@app.route('/api/tests/schedule', methods=['POST'])
@requires_auth('user')
def schedule_test():
    """Schedule a test"""
    job = request.get_json()
    job_id = uuid.uuid1()
    job['test_id'] = str(job_id)
    job['user'] = get_user_id()
    db.schedule_test(test_id=job_id, user=job['user'], 
                     cluster=job['cluster'], test_definition=job)
    return jsonify({'success':True, 'url':'/tests/id/{test_id}'.format(test_id=job['test_id'])})

@app.route('/api/tests/cancel', methods=['POST'])
@requires_auth('user')
def cancel_test():
    """Cancel a scheduled test"""
    test_id = request.form['test_id']
    if user_in_role('admin'):
        db.update_test_status(test_id, 'cancelled')
    else:
        # Check if the test is owned by the user:
        test = db.get_test(test_id)
        if test['user'] == get_user_id():
            db.update_test_status(test_id, 'cancelled')
        else:
            return make_response(jsonify({'error':'Access Denied to modify test {test_id}'
                            .format(test_id=test_id)}), 401)
    return jsonify({'success':'Test cancelled'})
    
@requires_auth('user')
@app.route('/api/tests/id/<test_id>')
def get_test(test_id):
    """Retrieve the definition for a scheduled test"""
    try:
        test = db.get_test(test_id)
        log.debug(test)
        return jsonify(test)
    except UnknownTestError:
        return make_response(jsonify({'error':'Unknown Test {test_id}.'.format(test_id=test_id)}), 404)

@requires_auth('user')
@app.route('/api/clusters')
def get_clusters():
    """Retrieve information about available clusters"""
    clusters = db.get_cluster_names()
    return make_response(jsonify({'clusters':clusters}))


################################################################################
#### Websockets
################################################################################

@requires_auth('user')
@sockets.route('/api/console')
def console_messages(ws):
    """Receive console messages as they happen

    ZMQ message format:
     Console messages:
      console cluster_name {"job_id":"current_job_id", "msg":"message from console"}
     Control messages:
      Keep alive:
      The cluster is starting a job:
       console cluster_name {"job_id":"current_job_id", "ctl":"START"}
      The cluster finished a job:
       console cluster_name {"job_id":"current_job_id", "ctl":"DONE"}
      The cluster is not working on anything:
       console cluster_name {"ctl":"IDLE"}

    When forwarding messages to the websocket client, the "console cluster_name" 
    portion is dropped and just the JSON is sent.

    Websocket sends keepalive messages periodically:
     {"ctl":"KEEPALIVE"}

    """
    cluster_name = ws.receive()
    console_socket = console_subscribe(cluster_name)
    try:
        while True:
            try:
                data = console_socket.recv_string()
                data = data.lstrip("console {cluster_name} ".format(cluster_name=cluster_name))
                ws.send(data)
            except zmq.error.Again:
                # If we timeout from zmq, send a keep alive request to the
                # websocket client:
                ws.send('{"ctl":"KEEPALIVE"}')
                # The client websocket will send keepalive back:
                ws.receive()
            except zmq.error.ZMQError, e:

                if e.errno == zmq.POLLERR:
                    log.error(e)
                    # Interrupted zmq socket code, reinitialize:
                    # I get this when I resize my terminal.. WTF?
                    console_socket = setup_zmq()
    finally:
        log.error("Unsubscribing from zmq socket")
        console_socket.setsockopt_string(zmq.UNSUBSCRIBE, u'')



"""Backend API to communicate with test clusters"""

import json
import base64
import uuid
import os
from flask import Flask
import zmq
import hashlib
import cStringIO

from app import app, db, sockets
from cstar_perf.frontend.lib.crypto import APIKey, BadConfigFileException
from cstar_perf.frontend.lib.util import random_token
import cstar_perf.frontend.lib.socket_comms as socket_comms
from cstar_perf.frontend.lib.socket_comms import Command, Response, receive_data, UnauthenticatedError
from cstar_perf.frontend import SERVER_KEY_PATH
from notifications import console_publish

import logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('cstar_perf.cluster_api')

class BadResponseError(Exception):
    pass

class CloseConnectionException(Exception):
    pass


@sockets.route('/api/cluster_comms')
def cluster_comms(ws):
    """Websocket to communicate with the test clusters

    Commands are logical actions one end of the socket wishes the
    other end to take. Responses are follow ups to a command, which
    there can be multiple, back and forth between client and server
    until the receiving end marks a response as Done.

    Command structure:
     {type:'command',
      command_id:'some unique string for this command',
      message:'some message to the other end',
      action:'some action for the receiver to take',
      // Extra parameters:
      foo: ...,
      bar: ...,
     }

    Response structure:
     {type:'response',
      command_id:'the command id this is a response to',
      message:'some message to the other end',
      done: true/false (the responder considers the command complete)
      // Extra parameters:
      foo: ...,
      bar: ...,
     }

    Possible commands:
      * authenticate - server asks client to authenticate
      * get_work - client asks for a test
      * test_done - client is done with a test, and sending artifacts
      * cancel_test - server asks client to cancel test
      * shutdown - server asks client to shutdown service

    Protocol:
     Authentication:
      * client initiates connection to this server
      * server sends client a random challenge token
        {type:'command', command_id='zzzz', action:'authenticate', token:'xxxxxxx'}
      * client signs challenge token with it's private key ands sends the signature
        {type:'response', command_id='zzz', cluster:'bdplab', signature:'xxxxxxxx'}
      * server verifies the signature is against the token it sent and the public
        key it has on file for the cluster.
      * server sends a 'you are authenticated' response.
        {type:'response', command_id='zzz', authenticated: true, done:true}

     Task loop:
      * client sends a 'give me work' request.
        {type:'command', command_id='yyy', action:'get_work'}
      * server sends a 'ok, wait for work' response.
        {type:'response', command_id='yyy', action:'wait'}
      * server sends a single test to the cluster
        {type:'response', command_id='yyy', test:{...}}
      * client responds 'ok, received test' response
        {type:'response', command_id:'yyy', test_id:'xxxxxxx'}
      * server updates status of test to in_progress in database
        {type:'response', command_id:'yyy', message:'test_updated', done:true}
      * client sends artifacts via streaming protocol (See below)
      * client sends 'ok, test done, artifacts sent.' request.
        {type:'command', command_id:'llll', action:'test_done', test_id:'xxxxxxx'}
      * server updates status of test to completed
      * server sends a 'ok, test updated' response
        {type:'response', command_id:'llll', test_id:'xxxxxx', message='test_update', done:true}

     Streaming:
      protocol for streaming raw data: console output, binary artifacts etc.
      * Sending peer sends a "I'm going to send binary data to you" request:
        {type:'command', command_id='xxx', action:'stream', test_id='xxxxx', 
         kind:"[console|failure|chart|system_logs|stress_logs]", name='name', 
         eof='$$$EOF$$$', keepalive='$$$KEEPALIVE$$$'}
      * Receiving peer sends response indicating it's ready to receive the stream:
        {type:'response', command_id='xxx', action='ready'}
      * Peer starts sending arbitrary binary data messages.
      * The receiving peer reads binary data. If it encounters $$$KEEPALIVE$$$ as it's own message, it will 
        omit that data, as it's only meant to keep the socket open.
      * Once $$$EOF$$$ is seen by the receiving peer, in it's own message, the receiving peer can respond:
        {type:'response', command_id='xxx', message:'stream_received', done:true}

    """
    context = {'apikey': APIKey.load(SERVER_KEY_PATH),
               'cluster': None}

    def authenticate():
        token_to_sign = random_token()
        cmd = Command.new(ws, action='authenticate', token=token_to_sign)
        response = cmd.send()
        context['cluster'] = cluster = response['cluster']
        client_pubkey = db.get_pub_key(cluster)
        client_apikey = APIKey(client_pubkey['pubkey'])
        
        # Verify the client correctly signed the token:
        try:
            client_apikey.verify_message(token_to_sign, response.get('signature'))
        except:
            response.respond(message='Bad Signature of token for authentication', done=True)
            log.error('client provided bad signature for auth token')
            raise

        response.respond(authenticated=True, done=True)

        # Client will ask us to authenticate too:
        command = receive_data(ws)
        assert command.get('action') == 'authenticate'
        data = {'signature' :context['apikey'].sign_message(command['token'])}
        response = command.respond(**data)
        if response.get('authenticated') != True:
            raise UnauthenticatedError("Our peer could not validate our signed auth token")

    def get_work(command):
        # Mark any existing in_process jobs for this cluster as
        # failed. If the cluster is asking for new work, then these
        # got dropped on the floor:
        for test in db.get_in_progress_tests(context['cluster']):
            db.update_test_status(test['test_id'], 'failed')

        # Find the next test scheduled for the client's cluster:
        tests = db.get_scheduled_tests(context['cluster'], limit=1)
        if len(tests) > 0:
            test_id = tests[0]['test_id']
        else:
            # No tests are currently scheduled.
            # Register a zmq listener of notifications of incoming tests, with a timeout.
            # When we see any test scheduled notification for our cluster, redo the query.
            # If timeout reached, redo the query anyway in case we missed the notification.
            def setup_zmq():
                zmq_context = zmq.Context()
                zmq_socket = zmq_context.socket(zmq.SUB)
                zmq_socket.connect('tcp://127.0.0.1:5557')
                zmq_socket.setsockopt_string(
                    zmq.SUBSCRIBE, 
                    unicode('scheduled {cluster} '.format(cluster=context['cluster'])))
                zmq_socket.setsockopt(zmq.RCVTIMEO, 15000)
                return zmq_socket
            zmq_socket = setup_zmq()
            while True:
                try:
                    cluster, test_id = zmq_socket.recv_string().split()
                except zmq.error.Again:
                    pass
                except zmq.error.ZMQError, e:
                    if e.errno == zmq.POLLERR:
                        log.error(e)
                        # Interrupted zmq socket code, reinitialize:
                        # I get this when I resize my terminal.. WTF?
                        zmq_socket = setup_zmq()
                finally:
                    tests = db.get_scheduled_tests(context['cluster'], limit=1)
                    if len(tests) > 0:
                        test_id = tests[0]['test_id']
                        break
                    else:
                        # Send no-work-yet message:
                        console_publish(context['cluster'], {'ctl':'WAIT'})
                        command.respond(action='wait', follow_up=False)
        test = db.get_test(test_id)
        # Give the test to the client:
        response = command.respond(test=test)
        # Expect an prepared status message back:
        assert response['test_id'] == test['test_id'] and \
            response['status'] == 'prepared'
        # Update the test status:
        db.update_test_status(test['test_id'], 'in_progress')
        # Let the client know they can start it:
        response.respond(test_id=test['test_id'], status="in_progress", done=True)

    def test_done(command):
        """Receive completed test artifacts from client"""
        db.update_test_status(command['test_id'], command['status'])
        # Record test failure message, if any:
        if command['status'] == 'failed':
            msg = (command.get('message','') + "\n" + command.get('stacktrace','')).strip()
            db.update_test_artifact(command['test_id'], 'failure', msg)
        # Send response:
        command.respond(test_id=command['test_id'], message='test_update', done=True)

    def receive_stream(command):
        """Receive a stream of data"""
        command.respond(message="ready", follow_up=False)
        log.debug("Receving data stream ....")
        if command['kind'] == 'console':
            console_dir = os.path.join(os.path.expanduser("~"), ".cstar_perf", "console_out")
            try:
                os.makedirs(console_dir)
            except OSError:
                pass
            console = open(os.path.join(console_dir, command['test_id']), "w")
        tmp = cStringIO.StringIO()
        sha = hashlib.sha256()
        try:
            def frame_callback(frame):
                if command['kind'] == 'console':
                    console.write(frame)
                    console_publish(context['cluster'], {'job_id':command['test_id'], 'msg':frame})
                    console.flush()
                else:
                    console_publish(context['cluster'], {'job_id':command['test_id'], 'ctl':'IN_PROGRESS'})
                sha.update(frame)
                tmp.write(frame)
            socket_comms.receive_stream(ws, command, frame_callback)
            if command['kind'] == 'console':
                console.close()
            # TODO: confirm with the client that the sha is correct
            # before storing
        finally:
            # In the event of a socket error, we always want to commit
            # what we have of the artifact to the database. Better to
            # have something than nothing. It's the client's
            # responsibility to resend artifacts that failed.

            db.update_test_artifact(command['test_id'], command['kind'], tmp, command['name'])

        command.respond(message='stream_received', done=True, sha256=sha.hexdigest())
        
    # Client and Server both authenticate to eachother:
    authenticate()
    
    try:
        # Dispatch on client commands:
        while True:
            command = receive_data(ws)
            assert command['type'] == 'command'
            if command['action'] == 'get_work':
                console_publish(context['cluster'], {'ctl':'WAIT'})
                get_work(command)
            elif command['action'] == 'test_done':
                console_publish(context['cluster'], {'ctl':'DONE'})
                test_done(command)
            elif command['action'] == 'stream':
                receive_stream(command)
            elif command['action'] == 'good_bye':
                log.info("client said good_bye. Closing socket.")
                break
    finally:
        console_publish(context['cluster'], {'ctl':'GOODBYE'})



# Global list of available format parsers on your system
# mapped to the callable/Exception to parse a string into a dict
formats = {}

class MalformedJSON(Exception): pass

# json - simplejson or packaged json as a fallback
try:
    import simplejson
    formats['json'] = (simplejson.loads, simplejson.decoder.JSONDecodeError, MalformedJSON)
except ImportError:
    try:
        import json
        formats['json'] = (json.loads, ValueError, MalformedJSON)
    except ImportError:
        pass

import os
import sys
from optparse import OptionParser

from jinja2 import Environment, FileSystemLoader

def cli(opts, args):
    if args[1] == '-':
        data = sys.stdin.read()
    else:
        data = open(os.path.join(os.getcwd(), os.path.expanduser(args[1]))).read()

    try:
        data = formats['json'][0](data)
    except formats['json'][1]:
        raise formats['json'][2](u'%s ...' % data[:60])
        sys.exit(1)

    env = Environment(loader=FileSystemLoader(os.getcwd()))
    sys.stdout.write(env.get_template(args[0]).render(data))
    sys.exit(0)


def main():
    default_format = 'json'
    if default_format not in formats:
        default_format = sorted(formats.keys())[0]

    parser = OptionParser(usage="usage: %prog [options] <input template> <input data>")
    opts, args = parser.parse_args()

    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    # Without the second argv, assume they want to read from stdin
    if len(args) == 1:
        args.append('-')

    cli(opts, args)

if __name__ == "__main__":
    main()



import datetime as dt

def val_ago(value, unit="unit"):
    if value == 1:
        return "{} {} ago".format(value, unit)
    else:
        return "{} {}s ago".format(value, unit)

def human_date(dateval, nowfunc=dt.datetime.now):
    now = nowfunc()
    delta = now - dateval
    days = delta.days
    if days == 0:
        seconds = delta.seconds
        minutes = seconds / 60
        hours = minutes / 60
        if hours > 0:
            return val_ago(hours, unit="hour")
        if minutes > 0:
            return val_ago(minutes, unit="minute")
        return val_ago(seconds, unit="second")
    elif 0 < days < 7:
        return val_ago(days, unit="day")
    else:
        return dateval.strftime("%m/%d/%Y")

if __name__ == "__main__":
    jan1 = dt.datetime(2013, 1, 1)
    def test_case(expected, **kwargs):
        val = jan1 - dt.timedelta(**kwargs)
        human = human_date(val, nowfunc=lambda: jan1)
        assert human == expected, human
    test_case("1 day ago", days=1)
    test_case("2 days ago", days=2)
    test_case("5 seconds ago", seconds=5)
    test_case("2 minutes ago", seconds=60*2)
    test_case("3 hours ago", seconds=60*60*3)
    test_case("12/25/2012", days=7)



"""A collection of ZeroMQ servers

test_notification_service - send out notifications of test status changes
 - Registers a PULL socket that model.py sends notifications of tests to.
 - Registers a PUB socket that broadcasts notifications to cluster_api websocket subscribers.

console_monitor_service - monitor the console out of a cluster

"""

import zmq
from daemonize import Daemonize
import argparse
import logging
import traceback
import threading
import time
from collections import defaultdict, deque
from functools import partial
import json
import datetime

log = logging.getLogger(__name__)


TEST_NOTIFICATION_PORT_PUSH = 5556
TEST_NOTIFICATION_PORT_SUB = 5557
CONSOLE_MONITOR_PORT_PUSH = 5558
CONSOLE_MONITOR_PORT_SUB = 5559

def test_notification_service(port_pull=TEST_NOTIFICATION_PORT_PUSH, port_pub=TEST_NOTIFICATION_PORT_SUB, ip='127.0.0.1'):
    url_pull = "tcp://{ip}:{port_pull}".format(**locals())
    url_pub = "tcp://{ip}:{port_pub}".format(**locals())
    try:
        log.info('test_notification_service staring')
        log.info('test_notification_service pull url: {url_pull}'.format(**locals()))
        log.info('test_notification_service pub url: {url_pub}'.format(**locals()))
        context = zmq.Context()

        receiver = context.socket(zmq.PULL)
        receiver.bind(url_pull)

        publisher = context.socket(zmq.PUB)
        publisher.bind(url_pub)

        while True:
            data = receiver.recv_string()
            log.info('notification: {data}'.format(data=data))
            publisher.send_string(data)

    except Exception, e:
        # Log every error. If we're not running in the foreground, we
        # won't see the errrors any other way:
        log.error(traceback.format_exc())
        log.info("test_notification_service shutdown")

def console_monitor_service(port_pull=CONSOLE_MONITOR_PORT_PUSH, port_pub=CONSOLE_MONITOR_PORT_SUB, ip='127.0.0.1'):
    url_pull = "tcp://{ip}:{port_pull}".format(**locals())
    url_pub = "tcp://{ip}:{port_pub}".format(**locals())
    try:
        log.info('console_monitor_service staring')
        log.info('console_monitor_service pull url: {url_pull}'.format(**locals()))
        log.info('console_monitor_service pub url: {url_pub}'.format(**locals()))
        context = zmq.Context()

        receiver = context.socket(zmq.PULL)
        receiver.bind(url_pull)

        publisher = context.socket(zmq.XPUB)
        publisher.bind(url_pub)

        poller = zmq.Poller()
        poller.register(receiver, zmq.POLLIN)
        poller.register(publisher, zmq.POLLIN)

        # Cache the last 100 messages per cluster:
        cache = defaultdict(partial(deque, maxlen=100)) # cluster_name -> deque

        while True:
            events = dict(poller.poll(1000))
            
            if receiver in events:
                data = receiver.recv()
                topic, cluster, msg = data.split(' ', 2)
                cache[cluster].append(msg)
                # Mark message as realtime:
                msg = json.loads(msg)
                msg['realtime'] = True
                msg = json.dumps(msg)
                data = " ".join([topic, cluster, msg])
                log.debug("PUB - {msg}".format(msg=data))
                publisher.send(data)
            
            if publisher in events:
                event = publisher.recv()
                # Subscription events areone byte: 0=unsub or 1=sub,
                # followed by topic:
                if event[0] == b'\x01':
                    topic, cluster = event[1:].strip().split(" ")
                    log.debug("SUBSCRIBE - {sub}".format(sub=event[1:]))
                    if topic == 'console':
                        # Client subscribed, send out previous messages:
                        log.debug("Sending backlog:")
                        for msg in cache[cluster]:
                            # Mark messages as non-realtime:
                            data = json.loads(msg)
                            data['realtime'] = False
                            msg = json.dumps(data)
                            data = "console {cluster} {msg}".format(cluster=cluster, msg=msg)
                            log.debug(data)
                            publisher.send(data)
                elif event[0] == b'\x00':
                    log.debug("UNSUBSCRIBE - {sub}".format(sub=event[1:]))
                    

    except Exception, e:
        # Log every error. If we're not running in the foreground, we
        # won't see the errrors any other way:
        log.error(traceback.format_exc())
        log.info("console_monitor_service shutdown")

def multi_service():
    """Start all the services in separate threads"""
    threads = []
    for service in [test_notification_service, console_monitor_service]:
        threads.append(threading.Thread(target=service))
    for thread in threads:
        thread.daemon = True
        thread.start()
    while threading.active_count() > 0:
        try:
            time.sleep(0.1)
        except KeyboardInterrupt:
            exit()

def zmq_socket_subscribe(url, topic='', timeout=5000):
    zmq_context = zmq.Context()
    zmq_socket = zmq_context.socket(zmq.SUB)
    zmq_socket.connect(url)
    zmq_socket.setsockopt_string(
        zmq.SUBSCRIBE, 
        unicode(topic))
    # Timeout:
    zmq_socket.setsockopt(zmq.RCVTIMEO, timeout)
    return zmq_socket
    
def console_publish(cluster_name, data):
    """Publish a console message or control message

    cluster_name - the name of the cluster the data came from
    data - a dictionary containing the following:
       job_id - the job id the cluster is currently working on
       msg - a message shown on the console
       ctl - A control message indicating cluster status START, DONE, IDLE
    """
    zmq_context = zmq.Context()
    zmq_socket = zmq_context.socket(zmq.PUSH)
    zmq_socket.connect("tcp://127.0.0.1:{port}".format(port=CONSOLE_MONITOR_PORT_PUSH))
    if not data.has_key('timestamp'):
        data['timestamp'] = (datetime.datetime.utcnow() - datetime.datetime(1970,1,1)).total_seconds()
    zmq_socket.send_string("console {cluster_name} {data}".format(
        cluster_name=cluster_name, 
        data=json.dumps(data)))

def console_subscribe(cluster_name):
    return zmq_socket_subscribe(
        'tcp://localhost:{port}'.format(port=CONSOLE_MONITOR_PORT_SUB), 
        'console {cluster_name} '.format(cluster_name=cluster_name))



def main():
    parser = argparse.ArgumentParser(description='cstar_perf_notifications')
    parser.add_argument('-F', '--foreground', dest='foreground', 
                        action='store_true', help='Run in the foreground instead of daemonizing')
    parser.add_argument('--pid', default="/tmp/cstar_perf_notifications.pid", 
                        help='PID file for daemon', dest='pid')
    parser.add_argument('-l', '--log', default='/tmp/cstar_perf_notifications.log',
                        help='File to log to', dest='logfile')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Print log messages', dest='verbose')
    args = parser.parse_args()

    log.setLevel(logging.DEBUG)
    log.propagate = False
    fh = logging.FileHandler(args.logfile, "a")
    formatter = logging.Formatter("%(levelname)s:%(funcName)s:%(asctime) -8s %(message)s")
    fh.setFormatter(formatter)
    fh.setLevel(logging.DEBUG)
    log.addHandler(fh)
    if args.verbose:
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        sh.setLevel(logging.DEBUG)
        log.addHandler(sh)
    keep_fds = [fh.stream.fileno()]


    if args.foreground:
        multi_service()
    else:
        daemon = Daemonize(app="notifications", pid=args.pid, action=multi_service, keep_fds=keep_fds)
        daemon.start()


if __name__ == "__main__":
    main()



from urllib2 import urlopen, URLError
import datetime as dt
from pymongo import MongoClient

TEAM_NAME = "hacknode1"

def get_collection():
    return MongoClient()[TEAM_NAME].articles 

def top_articles():
    coll = get_collection()
    articles = coll.find()
    return list(articles)

def search_articles(query):
    print "Searching ->", query
    articles = coll.find({"title": 
                            {"$regex": query}
                         })
    return list(articles)

def insert_article(article):
    coll = get_collection()
    existing = coll.find_one({"link": article["link"]})
    if existing is not None:
        print "Found existing, explicit upvoting ->", existing
        coll.update({"link": existing["link"]},
                    {"$inc":
                        {"score": 5}
                    })
        return True
    else:
        article["score"] = 0
        article["date"] = dt.datetime.now()
        print "Inserting ->", article
        coll.insert(article)
        return True

def track_click(url):
    coll = get_collection()
    print "Tracking ->", url
    coll.update({"link": url}, 
                {"$inc": 
                    {"score": 1}
                })
    return True

def validate_submission(params):
    errors = {}
    def err(id, msg):
        errors[id] = msg
    title = params["title"]
    title = title.strip()
    if len(title) < 2:
        err("title", "title must be > 2 characters")
    if len(title) > 150:
        err("title", "title may not be > 150 characters")
    link = params["link"]
    link = link.strip()
    try:
        opened = urlopen(link)
        link = opened.geturl()
    except (URLError, ValueError):
        err("link", "link could not be reached")
    if len(errors) > 0:
        return (False, errors)
    else:
        return (True, errors)



from werkzeug.wrappers import Request, Response
from werkzeug.debug import DebuggedApplication

@Request.application
def app(request):
    raise ValueError("testing debugger")
    return Response("hello, world!")

app = DebuggedApplication(app, evalex=True)

from werkzeug.serving import run_simple
run_simple("localhost", 4000, app)



import argparse
import logging
import os
import sys
import re
import base64
import ConfigParser
import json
import uuid
import time
import yaml
import copy
import subprocess
import pexpect
import tempfile
import tarfile
import base64
import hashlib
import shutil
import traceback
from collections import namedtuple
from distutils.dir_util import mkpath

import ecdsa
from daemonize import Daemonize
import fabric.api, fabric.network
from websocket import WebSocket

from cstar_perf.frontend.lib.crypto import APIKey, BadConfigFileException
from cstar_perf.frontend.lib.util import random_token, timeout, TimeoutError, format_bytesize, cd
from cstar_perf.frontend.lib.socket_comms import Command, Response, CommandResponseBase, receive_data, UnauthenticatedError
from cstar_perf.tool.stress_compare import stress_compare

logging.basicConfig(level=logging.DEBUG)
logging.getLogger('paramiko').setLevel(logging.WARNING)
log = logging.getLogger('cstar_perf.client')

from cstar_perf.frontend import CLIENT_CONFIG_PATH, KEEPALIVE_MARKER, EOF_MARKER

class JobFailure(Exception):
    pass

class JobRunner(object):
    """Periodically requests jobs from the remote server. 
    Runs them. 
    Reports back."""

    def __init__(self, ws_endpoint):
        self.ws_endpoint = ws_endpoint
        config = ConfigParser.RawConfigParser()
        config.read(CLIENT_CONFIG_PATH)
        self.__cluster_name = config.get('cluster','name')
        self.__client_key = APIKey.load(key_type='client')
        self.__server_key = APIKey.load(key_type='server')

        # Keep track of whether the server is connected and on the
        # same page as the client:
        self.__server_synced = False

    def connect(self):
        """Connect to the endpoint and return the opened websocket"""
        log.debug("Connecting to {ws_endpoint} ...".format(ws_endpoint=self.ws_endpoint))
        ws = self.ws = WebSocket()
        ws.connect(self.ws_endpoint)
        # Alias websocket-client's recv command to receieve so that
        # it's compatible with gevent-websocket:
        ws.receive = ws.recv

        # Authenticate:
        cmd = receive_data(ws)
        self.__authenticate(cmd)
        self.__server_synced = True
        return ws

    def send(self, data_or_command_response, assertions={}):
        return self.__socket_comms('send', data_or_command_response, assertions)

    def receive(self, ws_or_command_response, assertions={}):
        return self.__socket_comms('recv', ws_or_command_response, assertions)

    def respond(self, command_response, assertions={}):
        return self.__socket_comms('respond', command_response, assertions)

    def __socket_comms(self, method, obj, assertions={}):
        """Receive from a websocket, Command, or Response object.

        This wrapper is used to track websocket connection state and
        recover appropriately if the socket dies.

        method - 'send', 'recv', or 'respond'

        obj - Data to operate on, method dependent.
            send - either a text string to send on the websocket, or a prepared Command / Response object.
            recv - either a websocket object to receive from, or a Command / Response object.
            respond - a Command / Response object
    
        assertions - the key/value pairs of assertions, if specified,
          must exist in the Response object received. (Not applicable if
          ws_or_command is a raw websocket.)

        If there is a problem with the socket, or if any of the
        assertions fail, or if any prior receive call failed this way,
        this method will update the connection state but return
        silently. This failed state is only cleared on connect()

        """
        # Default response data if there's a problem:
        if isinstance(obj, CommandResponseBase):
            data_or_response = Response(obj.ws, {'type':'response', 'command_id':obj['command_id']})
        else:
            data_or_response = ""

        # Try to communicate, unless we know the server has desynchronized:
        if self.__server_synced:
            try:
                if isinstance(obj, CommandResponseBase) or isinstance(obj, WebSocket):
                    data_or_response = getattr(obj, method)()
                elif method == 'send':
                    # Assume obj is a peice of data to send on the raw websocket:
                    data_or_response = self.ws.send(obj)
                if len(assertions) > 0:
                    assert isinstance(data_or_response, CommandResponseBase)
                    assertions = set(assertions.items())
                    response_set = set(data_or_response.items())
                    missing_set = assertions.difference(response_set)
                    if len(missing_set) > 0:
                        raise AssertionError(
                            "missing fields in response: {missing_set}".format(
                                missing_set=missing_set))
            except Exception, e:
                log.error("Server desynchronized: {e}".format(e=e))
                log.error(traceback.format_exc(e))
                log.warn("Can't use the websocket anymore. I'll finish this job and then disconnect.")
                self.__server_synced = False


        return data_or_response

    def run(self):
        """Run a job, collect artifacts, send them to the server"""

        try:
            os.makedirs(os.path.join(os.path.expanduser("~"),'.cstar_perf','jobs'))
        except OSError:
            pass

        ws = self.connect()

        # Find old jobs, find their status, update the server
        self.recover_jobs()

        # Get work task loop:
        while True:
            # - Get next job.
            log.debug("Asking for a new job...")
            job = self.__get_work()
            log.debug("Got job: {job}".format(job=job))

            status_to_submit = "completed"
            message = None
            stacktrace = None
            try:
                self.perform_job(job, ws)
            except Exception, e:
                message = e.message
                stacktrace = traceback.format_exc(e)
                log.error(stacktrace)
                status_to_submit = "failed"

            # check if the server desynchronized (socket died or got
            # invalid response):
            if self.__server_synced == False:
                # Break out of this job loop. This will force a
                # disconnect and the job data will be resent upon
                # reconnection (self.recover_jobs())
                log.error("Disconnecting from server due to websocket desynchronization.")
                job_dir = os.path.join(os.path.expanduser("~"),'.cstar_perf','jobs',job['test_id'])
                log.error("Last job directory: {job_dir}".format(job_dir=job_dir))
                # Save any failure messages to the job directory to
                # send to the server later:
                with open(os.path.join(job_dir, 'failure.json'), 'w') as f:
                    json.dump({"message":message, "stacktrace":stacktrace}, f)
                break

            self.__job_done(job['test_id'], status=status_to_submit, message=message, stacktrace=stacktrace)
            
    def perform_job(self, job, ws):
        """Perform a job the server gave us, stream output and artifacts to the given websocket."""
        job = copy.deepcopy(job['test_definition'])
        # Cleanup the job structure according to what stress_compare needs:
        for revision in job['revisions']:
            cass_yaml = yaml.load(revision['yaml'])
            if cass_yaml is not None:
                if type(cass_yaml) is not dict:
                    raise JobFailure('Invalid yaml, was expecting a dictionary: {cass_yaml}'.format(cass_yaml=cass_yaml))
                revision.update(cass_yaml)
            del revision['yaml']
            if revision['options'] is not None:
                revision.update(revision['options'])
            del revision['options']
        for operation in job['operations']:
            operation['type'] = operation['operation']
            del operation['operation']

        job_dir = os.path.join(os.path.expanduser('~'),'.cstar_perf','jobs',job['test_id'])
        mkpath(job_dir)
        stats_path = os.path.join(job_dir,'stats.{test_id}.json'.format(test_id=job['test_id']))
        stress_log_path = os.path.join(job_dir,'stress_compare.{test_id}.log'.format(test_id=job['test_id']))

        stress_json = json.dumps(dict(revisions=job['revisions'],
                                      operations=job['operations'],
                                      title=job['title'],
                                      log=stats_path))

        # Create a temporary location to store the stress_compare json file:
        stress_json_path = os.path.join(job_dir, 'test.{test_id}.json'.format(test_id=job['test_id']))
        with open(stress_json_path, 'w') as f:
            f.write(stress_json)

        # Inform the server we will be streaming the console output to them:
        command = Command.new(self.ws, action='stream', test_id=job['test_id'], 
                              kind='console', name='console_out', eof=EOF_MARKER, keepalive=KEEPALIVE_MARKER)
        response = self.send(command, assertions={'message':'ready'})

        # Run stress_compare in a separate process, collecting the
        # output as an artifact:
        try:
            # Run stress_compare with pexpect. subprocess.Popen didn't
            # work due to some kind of tty issue when invoking
            # nodetool.
            stress_proc = pexpect.spawn('cstar_perf_stress {stress_json_path}'.format(stress_json_path=stress_json_path), timeout=None)
            with open(stress_log_path, 'w') as stress_log:
                while True:
                    try:
                        with timeout(25):
                            line = stress_proc.readline()
                            if line == '':
                                break
                            stress_log.write(line)
                            sys.stdout.write(line)
                            self.send(line)
                    except TimeoutError:
                        self.send(KEEPALIVE_MARKER)
        finally:
            self.send(EOF_MARKER)

        response = self.receive(response, assertions={'message':'stream_received', 'done':True})

        # Find the log tarball for each revision by introspecting the stats json:
        system_logs = []
        log_dir = os.path.join(os.path.expanduser("~"), '.cstar_perf','logs')
        
        with open(stats_path) as stats:
            stats = json.loads(stats.read())
            for rev in stats['revisions']:
                system_logs.append(os.path.join(log_dir, "{name}.tar.gz".format(name=rev['last_log'])))
        # Make a new tarball containing all the revision logs:
        tmptardir = tempfile.mkdtemp()
        try:
            job_log_dir = os.path.join(tmptardir, 'cassandra_logs.{test_id}'.format(test_id=job['test_id']))
            os.mkdir(job_log_dir)
            for x, syslog in enumerate(system_logs, 1):
                with tarfile.open(syslog) as tar:
                    tar.extractall(job_log_dir)
                    os.rename(os.path.join(job_log_dir, tar.getnames()[0]), os.path.join(job_log_dir, 'revision_{x:02d}'.format(x=x)))
            system_logs_path = os.path.join(job_dir, 'cassandra_logs.{test_id}.tar.gz'.format(test_id=job['test_id']))
            with tarfile.open(system_logs_path, 'w:gz') as tar:
                with cd(tmptardir):
                    tar.add('cassandra_logs.{test_id}'.format(test_id=job['test_id']))
            assert os.path.exists(system_logs_path)
        finally:
            shutil.rmtree(tmptardir)

        ## Stream artifacts
        ## Write final job status to 0.job_status file
        final_status = 'local_complete'
        try:
            # Stream artifacts:
            self.stream_artifacts(job['test_id'])
            if self.__server_synced:
                final_status = 'server_complete'

            # Spot check stats to ensure it has the data it should
            # contain. Raises JobFailure if something's amiss.
            try:
                self.__spot_check_stats(job, stats_path)
            except JobFailure, e:
                if final_status == 'server_complete':
                    final_status = 'server_fail'
                else:
                    final_status = 'local_fail'
                raise
        finally:
            with open(os.path.join(job_dir,'0.job_status'), 'w') as f:
                f.write(final_status)
            

    def stream_artifacts(self, job_id):
        """Stream all job artifacts

        Artifacts this looks for:
          console     - stress_compare console logs
            We already stream console logs during the job itself, but
            these can get interrupted, so, better to send it again.
          stats       - stress statistics (intervals, aggregates) JSON
          system_logs - cassandra logs

        returns a namedtuple of sent, failed to transmit, or missing artifacts.
        """
        streamed = []
        failed = []
        missing = []
        job_dir = os.path.join(os.path.expanduser("~"), ".cstar_perf","jobs",job_id)
        
        def stream(kind, name_pattern, binary):
            name = name_pattern.format(job_id=job_id)
            path = os.path.join(job_dir, name)
            if os.path.isfile(path):
                self.stream_artifact(job_id, kind, name, path, binary)
                if self.__server_synced:
                    streamed.append(kind)
                else:
                    failed.append(kind)
            else:
                missing.append(kind)

        for kind, pattern, binary in (
                ('console', 'stress_compare.{job_id}.log', False),
                ('stats', 'stats.{job_id}.json', False),
                ('system_logs', 'cassandra_logs.{job_id}.tar.gz', True)):
            stream(kind, pattern, binary)

        return namedtuple('StreamedArtifacts', 'streamed failed missing')(streamed, failed, missing)

    def stream_artifact(self, job_id, kind, name, path, binary=False):
        """Stream job artifact to server"""
        # Inform the server we will be streaming an artifact:
        command = Command.new(self.ws, action='stream', test_id=job_id, 
                              kind=kind, name=name, eof=EOF_MARKER, keepalive=KEEPALIVE_MARKER, binary=binary)
        response = self.send(command, assertions={'message':'ready'})

        fsize = format_bytesize(os.stat(path).st_size)
        with open(path) as f:
            log.info('Streaming {name} - {path} ({fsize})'.format(name=name, path=path, fsize=fsize))
            while True:
                data = f.read(512)
                if data == '':
                    break
                if binary:
                    data = base64.b64encode(data)
                self.send(data)
        if binary:
            self.send(base64.b64encode(EOF_MARKER))
        else:
            self.send(EOF_MARKER)

        response = self.receive(response, assertions={'message':'stream_received', 'done':True})

    def recover_jobs(self):
        """Find old jobs that are still on this machine and update the server on their state.

        This is used to cleanup after network or procedural failures.
        """
        # Iterate over every directory in ~/.cstar_perf/jobs
        log.info("Looking for old jobs that did not get sent to the server ...")
        for job_dir in os.listdir(os.path.join(os.path.expanduser("~"),'.cstar_perf', 'jobs')):
            job_dir = os.path.join(os.path.expanduser("~"),'.cstar_perf', 'jobs', job_dir)
            if not os.path.isdir(job_dir):
                continue
            job_id = os.path.split(job_dir)[-1]
            # Look for a file called 0.job_status which contains
            # hints as to the last step taken.
            test_status = 'local_fail'
            if os.path.isfile(os.path.join(job_dir, '0.job_status')):
                with open(os.path.join(job_dir, '0.job_status')) as f:
                    test_status = f.read().strip()
            if test_status.startswith('server'):
                # We uploaded all artifacts and the server has
                # marked with the final status. We can just delete
                # this directory:
                ### shutil.rmtree(job_dir)
                pass
            elif test_status == 'local_complete':
                # This job completed successfully, but it has not
                # been uploaded to the server, or there was a
                # problem uploading to the server. We should try
                # again:
                self.stream_artifacts(job_id)
                if self.__server_synced:
                    self.__job_done(job_id, status='completed')
                    with open(os.path.join(job_dir, '0.job_status'),'w') as f:
                        f.write('server_complete')                    
            else:
                # local_fail
                # This test did not get to local_complete status, so
                # we should upload whatever artifacts we have, and
                # tell the server the test failed
                self.stream_artifacts(job_id)
                if self.__server_synced:
                    failure_json = os.path.join(job_dir, 'failure.json')
                    failures = {}
                    if os.path.exists(failure_json):
                        with open(failure_json) as f:
                            failures=json.load(f)
                    self.__job_done(job_id, status='failed', **failures)
                    with open(os.path.join(job_dir, '0.job_status'),'w') as f:
                        f.write('server_fail')

            if not self.__server_synced:
                raise JobFailure("Server desynchronized while we were trying to send an old job. We'll try again later.")

    def __spot_check_stats(self, job, stats_path):
        """Spot check stats to ensure it has the data it should contain"""
        try:
            with open(stats_path) as stats:
                stats = json.loads(stats.read())
                for rev in job['revisions']:
                    for op_num, op in enumerate(job['operations']):
                        assert stats['stats'][op_num]['type'] == op['type']
                        assert stats['stats'][op_num]['command'].startswith(op['command'])
                        if op['type'] == 'stress':
                            assert len(stats['stats'][op_num]['intervals']) > 0
        except Exception, e:
            message = e.message
            stacktrace = traceback.format_exc(e)
            raise JobFailure("job stats is incomplete. message={message}\n{stacktrace}".format(message=message, stacktrace=stacktrace))


    def __get_work(self):
        """Ask the server for work"""
        command = Command.new(self.ws, action='get_work')
        response = command.send()
        while True:
            # We either got a job, or we received a wait request:
            if response.get('action') == 'wait':
                response = response.receive()
                continue
            elif response.has_key('test'):
                break
            else:
                raise AssertionError(
                    'Response was neither a wait action, nor contained '
                    'any test for us to run: {response}'.format(response=response))
        job = response['test']
        test_id = job['test_id']
        response = response.respond(test_id=test_id, status='prepared')
        assert response['status'] == 'in_progress'
        return job

    def __job_done(self, job_id, status='completed', message=None, stacktrace=None):
        """Tell the server we're done with a job, and give it the test artifacts"""
        ##{type:'command', command_id:'llll', action:'test_done', test_id:'xxxxxxx'}
        command = Command.new(self.ws, action="test_done", test_id=job_id, status=status)
        if message is not None:
            command['message'] = message
        if stacktrace is not None:
            command['stacktrace'] = stacktrace
        log.debug("Sending job completion message for {test_id} ...".format(test_id=job_id))
        response = command.send()
        ##{type:'response', command_id:'llll', test_id:'xxxxxx', message='test_update', done:true}
        assert response['test_id'] == job_id
        assert response['message'] == 'test_update'
        assert response['done'] == True
        log.debug("Server confirms job {test_id} is complete.".format(test_id=job_id))

    def __good_bye(self):
        """Tell the server we're disconnecting"""
        ##{type:'command', command_id:'llll', action:'good_bye'}
        command = Command.new(self.ws, action="good_bye")
        log.debug("Sending goodbye message to server..")
        command.send(await_response=False)

    def __authenticate(self, command):
        """Sign the token the server asked us to sign.
        Send it back.
        Give the server a token of our own to sign.
        Verify it."""
        assert command.get('action') == 'authenticate'
        data = {'signature' :self.__client_key.sign_message(command['token']),
                'cluster': self.__cluster_name}
        response = command.respond(**data)
        if response.get('authenticated') != True:
            raise UnauthenticatedError("Our peer could not validate our signed auth token")
        # cool, the server authenticated us, now we need to
        # authenticate the server:
        token = random_token()
        cmd = Command.new(self.ws, action='authenticate', token=token)
        response = cmd.send()
        signature = response['signature']
        # Verify the signature, raises BadSignatureError if it fails:
        try:
            self.__server_key.verify_message(token, signature)
        except:
            response.respond(message='Bad Signature of token for authentication', done=True)
            log.error('server provided bad signature for auth token')
            raise
        response.respond(authenticated=True, done=True)

def create_credentials():
    """Create ecdsa keypair for authenticating with server. Save these to
    a config file in the home directory."""
    print("Config file is : {config_path}".format(config_path=CLIENT_CONFIG_PATH))
    # Check for cluster name:
    config = ConfigParser.RawConfigParser()
    config.read(CLIENT_CONFIG_PATH)
    if not config.has_section("cluster"):
        config.add_section("cluster")
    if config.has_option('cluster','name'):
        cluster_name = config.get('cluster','name')
    else:
        while True:
            cluster_name = raw_input('Enter a name for this cluster: ')
            if not re.match(r'^[a-zA-Z0-9_]+$', cluster_name):
                print("Cluster name must be of the characters a-z, A-Z, 0-9, and _")
                continue
            break
        config.set('cluster','name', cluster_name)
        with open(CLIENT_CONFIG_PATH, "wb") as f:
            config.write(f)
        os.chmod(CLIENT_CONFIG_PATH, 0600)
    print("Cluster name is: {cluster_name}\n".format(cluster_name=cluster_name))

    # Check for existing client key:
    try:
        # Load existing config:
        apikey = APIKey.load()
    except BadConfigFileException:
        apikey = APIKey.new()
        apikey.save()

    print("Your public key is: {key}\n".format(key=apikey.get_pub_key()))

    # Check for existing server key:    
    try:
        # Load existing config:
        apikey = APIKey.load(key_type='server')
    except BadConfigFileException:
        server_key = raw_input("Input the server's public key:")
        try:
            apikey = APIKey(server_key)
        except:
            print("Invalid server key, does not decode to a valid ecdsa key.")
            exit(1)
        # Verify server key:
        verify_code = raw_input("Input the server verify code: ")
        token, sig = base64.decodestring(verify_code).split("|")
        apikey.verify_message(token, sig)
        apikey.save(key_type='server')
    
    print("Server public key is: {key}".format(key=apikey.get_pub_key()))

def main():
    parser = argparse.ArgumentParser(description='cstar_perf job client')
    parser.add_argument('-s', '--server', default='ws://localhost:8000/api/cluster_comms',
                        help='Server endpoint', dest='server')
    parser.add_argument('--get-credentials', dest='gen_credentials',
                        action='store_true', help='Get and/or create ECDSA key for signing requests.')

    args = parser.parse_args()

    if args.gen_credentials:
        create_credentials()
        return

    if not os.path.exists(CLIENT_CONFIG_PATH):
        print("Config file not found. run with --get-credentials to create them first.")
        exit(1)

    job_runner = JobRunner(args.server)
    job_runner.run()

if __name__ == "__main__":
    main()



import uuid
import json
import logging
import os

from flask import ( Flask, render_template, request, redirect, abort,
                    jsonify, make_response, session)
from flask.ext.script import Manager
from flask_sockets import Sockets

from model import Model, UnknownUserError, UnknownTestError
from util import csrf_protect_app, load_app_config
from cstar_perf.frontend.lib.util import random_token

logging.basicConfig(level=logging.DEBUG)
logging.getLogger('geventwebsocket').setLevel(logging.INFO)
log = logging.getLogger('cstar_perf')


app = Flask(__name__, static_folder="../static", static_url_path="/static")
app_config = load_app_config()
app.secret_key = app_config.get('server','app_secret')
app.debug = True
manager = Manager(app)
csrf_protect_app(app)
sockets = Sockets(app)

### Cassandra backend model:
db = Model(email_notifications=app_config.has_section('smtp'))

### Main application controllers:
import controllers

### Backend API controllers:
import cluster_api

if __name__ == "__main__":
    manager.run()
    



import json
import uuid
import logging
import Queue
import threading
import base64

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('cstar_perf.socket_comms')
from flask.json import JSONEncoder
from websocket import WebSocket

from cstar_perf.frontend import KEEPALIVE_MARKER, EOF_MARKER

class UnauthenticatedError(Exception):
    pass

class FaultSuppressedWebSocket(WebSocket):
    """A WebSocket that suppresses send() errors when not connected

    This does not suppress recv() errors. You need to catch socket.error exceptions yourself.
    """
    def send(self, payload, opcode=1):
        if self.connected:
            try:
                WebSocket.send(self, payload, opcode)
            except socket.error:
                pass

class CommandResponseBase(dict):
    """Base class for Command and Response classes"""
    def __init__(self, ws, data):
        self.update(data)
        self.ws = ws
        self._verify_type()
    
    def _verify_type(self):
        raise TypeError("Each subclass of CommandResponseBase must implement it's own _verify_type method")

    def respond(self, follow_up=True, **kwargs):
        """Send a response back to the peer
        
        Normally, a response requires another response from the peer until done=True. If follow_up==False, then we will not look for a response yet. This is used in action='wait' responses where we are going to send multiple responses in a row to the peer.
        """
        if self['type'] == 'response':
            assert self['done'] == False, "Can't respond to a response that is already marked done."
        data = {'command_id': self['command_id'], 'type':'response'}
        data.update(kwargs)
        if not data.has_key('done'):
            data['done'] = False
        data_str = json.dumps(data, cls=JSONEncoder)
        log.debug("Sending response : {data}".format(data=data_str))
        self.ws.send(data_str)
        if data['done'] == False and follow_up:
            # We are still expecting a response to our response:
            return self.receive()

    def receive(self):
        """Await for further response from our peer"""
        response = json.loads(self.ws.receive())
        log.debug("Received response : {response}".format(response=response))
        if response.get('type') == 'response':
            if response.get('command_id') == self['command_id']:
                return Response(self.ws, response)
            else:
                raise AssertionError("Unexpected response id in : {stuff}".format(stuff=response))
        else:
            raise AssertionError("Was expecting a response, instead got {stuff}".format(stuff=response))

    def recv(self):
        """Alias for receieve() compatible with WebSocket interface"""
        return self.receive()
        

class Command(CommandResponseBase):
    """A command our other peer asked us to do, or we want them to do"""
    def _verify_type(self):
        assert self.get('type') == 'command' and self.has_key('command_id') and self.has_key('action')

    @staticmethod
    def new(ws, **kwargs):
        """Create a new Command to be sent to our peer.
        This differs from the __init__ - that is reserved for instantiating commands originating FROM our peer."""
        assert 'command_id' not in kwargs, "New commands should not be given a command_id yet"
        assert 'type' not in kwargs, "New commands don't need to be passed a type parameter, one will be assigned."
        data = {'command_id': str(uuid.uuid1()),
                'type': 'command'}
        data.update(kwargs)
        cmd = Command(ws, data)
        cmd.new_command = True
        return cmd

    def send(self, await_response=True):
        """Send this command to our peer for them to execute"""
        #In order to send a command, there needs to be a
        #new_command==True flag on this command. Otherwise, this
        #command will be assumed to have originated from our peer and
        #should not be sent.
        if not getattr(self, 'new_command', False):
            raise AssertionError("Cannot send command that is not marked as new_command")
        data = json.dumps(self, cls=JSONEncoder)
        log.debug("Sending command : {data}".format(data=data))
        self.ws.send(data)
        if not await_response:
            return
        # Wait for the response:
        response = json.loads(self.ws.receive())
        log.debug("Received response : {response}".format(response=response))
        if response.get('type') == 'response':
            if response.get('command_id') == self['command_id']:
                return Response(self.ws, response)
            else:
                raise AssertionError("Unexpected response id in : {stuff}".format(stuff=response))
        else:
            raise AssertionError("Was expecting a response, instead got {stuff}".format(stuff=response))



class Response(CommandResponseBase):
    """A response to a Command, or a follow up response to a Response"""
    def _verify_type(self):
        assert self.get('type') == 'response' and self.has_key('command_id')

def receive_data(ws):
    data = json.loads(ws.receive())
    if data.get('type') == 'command':
        log.debug("Received command : {response}".format(response=data))
        return Command(ws, data)
    elif data.get('type') == 'response':
        log.debug("Received response : {response}".format(response=data))
        return Response(ws, data)
    else:
        raise ValueError('Unknown response type: {rtype}'.format(rtype=data.get('type')))

def receive_stream(ws, command, frame_callback):
    """Receive streaming binary data
    
    ws - the open websocket
    command - the Command object that issued the stream action. This will include the following things:
       - keepalive - denotes a frame that should be ignored, just for keeping the connection alive.
       - eof - denotes a frame that marks the end of the stream. This method will return True when it encounters this.
    frame_callback is a function to call on each non keepalive, non eof, frame. It takes a single argument for the frame data.
    """
    binary = command.get('binary', False)
    while True:
        data = ws.receive()
        if binary:
            data = base64.b64decode(data)
        if data == command['eof']:
            break
        elif data == command['keepalive']:
            pass
        else:
            frame_callback(data)



from filters import val_ago, human_date
from flask import render_template, Flask
import datetime as dt

app = Flask(__name__)

def top_articles():
    articles = [
        {"title": "Google", "score": 150, "link": "http://google.com"},
        {"title": "Yahoo", "score": 75, "link": "http://yahoo.com"},
        {"title": "Bing", "score": 50, "link": "http://bing.com"}
    ]
    return articles

@app.route('/')
def index():
    articles = top_articles()
    return render_template("index.jinja2.html", rows=articles)

@app.template_filter()
def seconds_ago(val):
    return val_ago(val, unit="second")

@app.route('/experiment')
def experiment():
    return render_template('seconds.jinja2.html',
                           seconds=range(60))

app.add_template_filter(human_date)

@app.route('/datetest')
def datetest():
    now = dt.datetime.now()
    deltas = [
        dt.timedelta(seconds=5),
        dt.timedelta(seconds=60*60),
        dt.timedelta(days=5),
        dt.timedelta(days=60)
    ]
    dates = [(now - delta)
                for delta in deltas]
    return render_template('dates.jinja2.html',
                            dates=dates)

if __name__ == "__main__":
    app.run(debug=True)



import os
import base64
import ConfigParser
import uuid
from distutils.dir_util import mkpath

import ecdsa

SERVER_KEY_PATH = os.path.join(os.path.expanduser("~"),'.cstar_perf', 'server.conf')

class NoPrivateKeyException(Exception):
    pass

class BadConfigFileException(Exception):
    pass


class APIKey(object):
    """Object representing the client or server api keys
    
    wraps ecdsa signing and verifying keys and returns/accepts base64 encoded messages.

    >>> apikey1 = APIKey.new()
    >>> sig1 = apikey1.sign_message('hello, world!')
    >>> apikey1.verify_message('hello, world!', sig1)
    True
    >>> apikey2 = APIKey(apikey1.get_pub_key())
    >>> apikey2.verify_message('hello, world!', sig1)
    True
    """
    def __init__(self, pubkey=None, privkey=None):
        if pubkey is None and privkey is None:
            raise AssertionError("Both privkey and pubkey are None")
        if pubkey is not None:
            self.__verifying_key = ecdsa.VerifyingKey.from_string(base64.decodestring(pubkey))
        if privkey is not None:
            self.__signing_key = ecdsa.SigningKey.from_string(base64.decodestring(privkey))
            if pubkey and self.__signing_key.get_verifying_key().to_string() != \
               self.__verifying_key.to_string():
                raise AssertionError('Public key does not match supplied Private key')
            else:
                self.__verifying_key = self.__signing_key.get_verifying_key()

    def get_pub_key(self):
        try:
            return base64.encodestring(self.__signing_key.get_verifying_key().to_string()).strip()
        except AttributeError:
            return base64.encodestring(self.__verifying_key.to_string()).strip()

    def verify_message(self, message, signature):
        signature = base64.decodestring(signature)
        return self.__verifying_key.verify(signature, message)
            
    def sign_message(self, message):
        try:
            return base64.encodestring(self.__signing_key.sign(message)).strip()
        except AttributeError:
            raise NoPrivateKeyException('This APIKey object has no private key, and cannot sign messages.')

    def save(self, config_path=None, key_type='client'):
        """Save APIKey to config file"""
        if not config_path:
            config_path = os.path.join(os.path.expanduser("~"), ".cstar_perf","client.conf")
        config = ConfigParser.RawConfigParser()
        config.read(config_path)
        if not config.has_section("credentials"):
            config.add_section("credentials")
        if key_type == 'client':
            config.set('credentials', 'priv_key', base64.encodestring(self.__signing_key.to_string()))
        elif key_type == 'server':
            config.set('credentials', 'server_key', base64.encodestring(self.__verifying_key.to_string()))
        else:
            raise AssertionError('Unknown key_type: {key_type}'.format(key_type=key_type))
        mkpath(os.path.dirname(config_path))
        with open(config_path, "wb") as f:
            config.write(f)
        os.chmod(config_path, 0600)

    @staticmethod
    def new():
        """Create new APIKey, generating a fresh ecdsa key"""
        sk = ecdsa.SigningKey.generate()
        return APIKey(privkey=base64.encodestring(sk.to_string()))
        
    @staticmethod
    def load(config_path=None, key_type='client'):
        """Load APIKey from config file

        config_path - the full path to the config file
        key_type - 'client' or 'server'"""
        
        if not config_path:
            config_path = os.path.join(os.path.expanduser("~"), ".cstar_perf", "client.conf")
        if not os.path.exists(config_path):
            raise BadConfigFileException('Config file does not exist: {config_path}'.
                                         format(config_path=config_path))
        config = ConfigParser.RawConfigParser()
        config.read(config_path)
        if key_type == 'client':
            try:
                privkey = config.get('credentials','priv_key')
            except:
                raise BadConfigFileException("Config file has no existing priv_key")
            apikey = APIKey(privkey=privkey)
        elif key_type == 'server':
            try:
                pubkey = config.get('credentials', 'server_key')
            except:
                raise BadConfigFileException("Config file has no existing server_key")
            apikey = APIKey(pubkey)

        return apikey

def generate_server_keys():
    """One time function to create and store server side keys"""
    try:
        apikey = APIKey.load(SERVER_KEY_PATH,
                             # Yes, client is correct here, because we
                             # want a signing key:
                             key_type='client')
        print("Server keys already exist in {config_path}".format(config_path=SERVER_KEY_PATH))
    except BadConfigFileException:
        apikey = APIKey.new()
        apikey.save(SERVER_KEY_PATH,
                    # Yes, client is correct here, because we are
                    # using a signing key:
                    key_type='client')
        print("New server keys saved to {config_path}".format(config_path=SERVER_KEY_PATH))

    print("Server public key: {pubkey}".format(pubkey=apikey.get_pub_key()))
    token = str(uuid.uuid1())
    sig = apikey.sign_message(token)
    verify_code = base64.b64encode(token + "|" + sig)
    print("Server verify code: {verify_code}".format(verify_code=verify_code))



# Server email notifications to users:
import smtplib
from email.mime.text import MIMEText
import os
from jinja2 import Template, Environment, PackageLoader
import ConfigParser
import logging
import urlparse

from cstar_perf.frontend import SERVER_CONFIG_PATH
from cstar_perf.frontend.server.util import load_app_config

app_config = load_app_config()
template_env = Environment(loader=PackageLoader('cstar_perf.frontend.server', os.path.join('templates','email')))

log = logging.getLogger('cstar_perf.model')

class Email(object):
    message_template = 'default.jinja2' # name of jinja2 template file in templates/email
    subject = "Subject goes here - this is a regular jinja template - you can use {{vars}}"
    required_args = None # list of argument names that must appear in __init__ kwargs
    _vars = {
        "base_url": app_config.get('server', 'url')
    }
    additional_vars = {}

    def __init__(self, recipients=[], **kwargs):
        self.config = self.__get_config()
        self.setup(recipients, **kwargs)

    def __get_config(self, path=SERVER_CONFIG_PATH):
        defaults = {
            'from': 'cstar_perf@localhost',
            'server': 'localhost',
            'ssl': 'False',
            'port': '587'
        }
        config = ConfigParser.ConfigParser(allow_no_value=True)
        config.add_section('smtp')
        for k,v in defaults.items():
            config.set('smtp',k,v)
        config.read(path)
        cfg = dict(config.items('smtp'))
        cfg['ssl'] = config.getboolean('smtp','ssl')
        cfg['authenticate'] = cfg.has_key('user')
        return cfg

    def setup(self, recipients=[], **kwargs):
        self.recipients = recipients
        if type(recipients) in [str, unicode]:
            self.recipients = [recipients]
        if self.required_args is not None:
            missing_args = ", ".join(set(self.required_args).difference(kwargs.keys()))
            if len(missing_args) > 0:
                raise ValueError("Missing required args: {missing_args}".format(**locals()))
        if self.message_template == Email.message_template:
            raise NotImplementedError("You must subclass the Email class, defining your own message template")
        if self.subject == Email.subject:
            raise NotImplementedError("You must subclass the Email class, defining your own message subject")
        self.subject = Template(self.subject).render(**kwargs)

        v = kwargs.copy()
        v.update(self._vars)
        v.update(self.additional_vars)

        self.body = template_env.get_template(self.message_template).render(**v)

    def send(self):
        s = smtplib.SMTP(self.config['server'], self.config['port'])
        s.ehlo()
        if self.config['ssl']:
            s.starttls()
            s.ehlo()
        if self.config['authenticate']:
            s.login(self.config['user'], self.config['pass'])
        msg = MIMEText(self.body)
        msg['Subject'] = self.subject
        msg['From'] = self.config['from']
        msg['To'] = ", ".join(self.recipients)
        msg['BCC'] = self.config.get('always_bcc', None)
        s.sendmail(self.config['from'], self.recipients, msg.as_string())
        log.info("Sent email to {recipients}".format(recipients=self.recipients))

class TestStatusUpdateEmail(Email):
    message_template = "test_status_update.jinja2"
    required_args = ('status','name','test_id')
    subject = "[cstar_perf] test {{status|upper}} - {{name}}"



### Queries this app has to answer:
###
### Tests:
###  * Schedule a test to run
###    - schedule_test(test_id, user, cluster, test_definition)
###  * Retrieve a test by id
###    - get_test(test_id)
###  * Retrieve tests that have not yet run, status=new
###    - get_scheduled_tests
###  * Retrieve tests that a user has:
###    - get_scheduled_tests_by_user
###    - get_
###  * Get the next test a given cluster should run.
###    - get_next_scheduled_test
###  * Report a test that begins to be run, status=in_progress.
###    - update_test_status
###  * Report a test that failed to run properly, status=failed.
###    - update_test_status
###  * Retrieve tests that have failed, status=failed.
###    - get_failed_tests
###  * Report a test as having been run successfully, record artifacts
###    - update_test_status
###    - add_test_artificat
###  * Retrieve tests that have been completed, status=completed, 
###    - get_completed_tests
###  * Get completed test artifacts,  logs, stats, graph url.
###    - get_test_artifacts
###
### Users:
###  * Retrieve user by id (email), including set of roles (user, admin)
###
### Clusters:
###  * Create a new cluster, name, number of nodes, description
###  * Deadman switch check in - Each cluster needs to call into the
###    server on set interval, recording the time, the job the cluster
###    is currently working on, and it's status (is something else
###    using the cluster?)


import cassandra
from cassandra.cluster import Cluster
import json
import uuid
import logging
import datetime
import zmq
from collections import namedtuple

from cstar_perf.frontend.lib.util import random_token, uuid_to_datetime
from cstar_perf.frontend.server.email_notifications import TestStatusUpdateEmail

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('cstar_perf.model')
logging.getLogger('cassandra').setLevel(logging.WARNING)


class UnknownUserError(Exception):
    pass
class UnknownTestError(Exception):
    pass
class APIKeyExistsError(Exception):
    pass
class UnknownAPIKeyError(Exception):
    pass
class NoTestsScheduledError(Exception):
    pass


class Model(object):

    statements = {
        'insert_test': "INSERT INTO tests (test_id, user, cluster, status, test_definition) VALUES (?, ?, ?, ?, ?);",
        'select_test': "SELECT * FROM tests WHERE test_id = ?;",
        'update_test_set_status': "UPDATE tests SET status = ? WHERE test_id = ?",
        'update_test_set_status_completed': "UPDATE tests SET status = ?, completed_date = ? WHERE test_id = ?",
        'insert_test_status': "INSERT INTO test_status (status, cluster, test_id, user, title) VALUES (?, ?, ?, ?, ?);",
        'select_test_status_asc': "SELECT * FROM test_status WHERE status = ? AND cluster = ? ORDER BY cluster DESC, test_id ASC LIMIT ?",
        'select_test_status_desc': "SELECT * FROM test_status WHERE status = ? AND cluster= ? ORDER BY cluster ASC, test_id DESC LIMIT ?",
        'select_test_status_by_user': "SELECT * FROM test_status WHERE status = ? AND user = ? LIMIT ?",
        'select_next_scheduled': "SELECT * FROM test_status WHERE status = 'scheduled' AND cluster = ? ORDER BY cluster DESC, test_id ASC LIMIT 1",
        'select_test_status_all': "SELECT * FROM test_status WHERE status = ? LIMIT ?",
        'delete_test_status': "DELETE FROM test_status WHERE status= ? AND cluster = ? AND test_id = ?",
        'select_clusters_name': "SELECT name from clusters;",
        'insert_clusters': "INSERT INTO clusters (name, num_nodes, description) VALUES (?, ?, ?)",
        'insert_user': "INSERT INTO users (user_id, full_name, roles) VALUES (?, ?, ?);",
        'select_user': "SELECT * FROM users WHERE user_id = ?;",
        'select_user_roles': "SELECT roles FROM users WHERE user_id = ?;",
        'update_test_artifact': "UPDATE test_artifacts SET description = ?, artifact = ? WHERE test_id = ? AND artifact_type = ?;",
        'select_test_artifacts_by_type': "SELECT artifact_type, description FROM test_artifacts WHERE test_id = ? AND artifact_type = ?",
        'select_test_artifacts_all': "SELECT artifact_type, description FROM test_artifacts WHERE test_id = ? ORDER BY artifact_type ASC",
        'select_test_artifact_data': "SELECT artifact, description FROM test_artifacts WHERE test_id = ? AND artifact_type = ? LIMIT 1",
        'insert_test_completed': "INSERT INTO tests_completed (status, completed_date, test_id, cluster, title, user) VALUES (?, ?, ?, ?, ?, ?);",
        'select_test_completed': "SELECT * FROM tests_completed LIMIT ?;",
        'select_api_pubkey': "SELECT * FROM api_pubkeys WHERE name = ? LIMIT 1",
        'insert_api_pubkey': "INSERT INTO api_pubkeys (name, user_type, pubkey) VALUES (?, ?, ?);"
    }

    def __init__(self, cluster=Cluster(['127.0.0.1']), keyspace='cstar_perf', email_notifications=False):
        """Instantiate DB model object for interacting with the C* backend.

        cluster - Python driver object for accessing Cassandra
        keyspace - the keyspace to use
        email_notifications - if True, perform email notifications for some actions. Defaults to False.
        """
        log.info("Initializing Model...")
        self.cluster = cluster
        self.keyspace = keyspace
        self.email_notifications = email_notifications
        self.__shared_session = self.get_session()
        ## Prepare statements:
        self.__prepared_statements = {}
        for name, stmt in Model.statements.items():
            #log.debug("Preparing statement: {stmt}".format(stmt=stmt))
            self.__prepared_statements[name] = self.get_session().prepare(stmt)

        ### ZeroMQ publisher for announcing jobs as they come in:
        zmq_context = zmq.Context()
        self.zmq_socket = zmq_context.socket(zmq.PUSH)
        self.zmq_socket.connect("tcp://127.0.0.1:5556")

        log.info("Model initialized")

    def get_session(self, shared=True):
        try:
            if shared:
                try:
                    session = self.__shared_session
                except AttributeError:
                    session = self.cluster.connect(self.keyspace)
            else:
                session = cluster.connect(self.keyspace)
        except cassandra.InvalidRequest, e:
            # Only attempt to create the schema if we get an error that it
            # doesn't exist:
            if "Keyspace '{ks}' does not exist".format(ks=self.keyspace) in e.message:
                self.__create_schema()
            session = self.cluster.connect(self.keyspace)
        return session
                    
    def __create_schema(self, replication_factor=1):
        session = self.cluster.connect()
        log.info("Creating new schema in keyspace : {ks}".format(ks=self.keyspace))
        session.execute("""CREATE KEYSPACE {ks} WITH replication = {{'class': 'SimpleStrategy', 
                        'replication_factor': {replication_factor}}}""".format(
                            ks=self.keyspace,
                            replication_factor=replication_factor
                        ))

        session = self.cluster.connect(self.keyspace)

        # All test tests indexed by id:
        session.execute("CREATE TABLE tests (test_id timeuuid PRIMARY KEY, user text, cluster text, status text, test_definition text, completed_date timeuuid);")
        # Tests listed by status, sorted by timestamp, in descending
        # order. Descending order because the completed status will have
        # the largest number. 'scheduled' status will want to be queried
        # in ASC order.
        session.execute("CREATE TABLE test_status (status text, test_id timeuuid, cluster text, user text, title text, PRIMARY KEY (status, cluster, test_id)) WITH CLUSTERING ORDER BY (cluster ASC, test_id DESC);")
        session.execute("CREATE INDEX ON test_status (user);")
        # A denormalized copy of test_status for the completed tests.
        # This makes a reverse querying of completed tests for the
        # main page doable:
        session.execute("CREATE TABLE tests_completed (status text, completed_date timeuuid, test_id timeuuid, cluster text, title text, user text, PRIMARY KEY (status, completed_date)) WITH CLUSTERING ORDER BY (completed_date DESC)")

        # Test artifacts
        session.execute("CREATE TABLE test_artifacts (test_id timeuuid, artifact_type text, description text, artifact blob, PRIMARY KEY (test_id, artifact_type));")

        # Cluster information
        session.execute("CREATE TABLE clusters (name text PRIMARY KEY, num_nodes int, description text)")
        
        #Users
        session.execute("CREATE TABLE users (user_id text PRIMARY KEY, full_name text, roles set <text>);")
        #session.execute("INSERT INTO users (user_id, full_name, roles) VALUES ('ryan@datastax.com', 'Ryan McGuire', {'user','admin'});")

        # API keys
        session.execute("CREATE TABLE api_pubkeys (name text PRIMARY KEY, user_type text, pubkey text)")


    ################################################################################
    #### Test Management:
    ################################################################################
    def schedule_test(self, test_id, user, cluster, test_definition):
        session = self.get_session()
        test_definition['test_id'] = str(test_id)
        test_json = json.dumps(test_definition)
        session.execute(self.__prepared_statements['insert_test'], (test_id, user, cluster, 'scheduled', test_json))
        session.execute(self.__prepared_statements['insert_test_status'], ('scheduled', cluster, test_id, user, test_definition['title']))
        self.zmq_socket.send_string("scheduled {cluster} {test_id}".format(**locals()))
        return test_id

    def get_test(self, test_id):
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        try:
            test = session.execute(self.__prepared_statements['select_test'], (test_id,))[0]
        except IndexError:
            raise UnknownTestError('Unknown test {test_id}'.format(test_id=test_id))
        test = self.__test_row_to_dict(test)
        return test

    def update_test_status(self, test_id, status):
        assert status in ('scheduled', 'in_progress', 'completed', 'cancelled', 'failed'), "{status} is not a valid test state".format(status=status)
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        test = self.get_test(test_id)
        # Update tests table:
        if status == "completed":
            completed_date = uuid.uuid1()
            session.execute(self.__prepared_statements['update_test_set_status_completed'], (status, completed_date, test_id ))
            # Add denormalized copy in tests_
            session.execute(self.__prepared_statements['insert_test_completed'], ("completed", completed_date, test_id, test['cluster'], test['test_definition']['title'], test['user']))
        else:
            session.execute(self.__prepared_statements['update_test_set_status'], (status, test_id))
        # Remove the old status from test_status:
        session.execute(self.__prepared_statements['delete_test_status'], (test['status'], test['cluster'], test_id))
        # Add the new status to test_status:
        session.execute(self.__prepared_statements['insert_test_status'], (status, test['cluster'], test_id, test['user'], test['test_definition']['title']))
        self.zmq_socket.send_string("{status} {cluster} {test_id}".format(status=status, cluster=test['cluster'], test_id=test_id))
        log.info("test status is: {status}".format(status=status))

        # Send the user an email when the status updates:
        if self.email_notifications and status not in ('scheduled','in_progress'):
            TestStatusUpdateEmail([test['user']], status=status, name=test['test_definition']['title'], 
                                  test_id=test_id).send()
        return namedtuple('TestStatus', 'test_id status')(test_id, status)

    def update_test_artifact(self, test_id, artifact_type, artifact, description=None):
        """Update an artifact blob

        artifact can be a string or a file-like object

        test_id,artifact_type are unique in the database. If the pair already exists, it will be overridden.
        """
        if hasattr(artifact, 'read'):
            f = artifact
            pos = f.tell()
            f.seek(0)
            artifact = f.read()
            f.seek(pos)
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        if not description:
            description = "Unknown artifact"
        artifact = artifact.encode("hex")
        session.execute(self.__prepared_statements['update_test_artifact'], (description, artifact, test_id, artifact_type))
        return test_id

    def get_test_artifact(self, test_id, artifact_type):
        """Retrieve one test artifact type"""
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        rows = session.execute(self.__prepared_statements['select_test_artifacts_by_type'], (test_id, artifact_type))
        return [r.__dict__ for r in rows][0]

    def get_test_artifacts(self, test_id, artifact_type=None):
        """Retrieve all test artifacts"""
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        rows = session.execute(self.__prepared_statements['select_test_artifacts_all'], (test_id,))
        return [r.__dict__ for r in rows]

    def get_test_artifact_data(self, test_id, artifact_type):
        """Get blob data from a specific artifact"""
        session = self.get_session()
        if not isinstance(test_id, uuid.UUID):
            test_id = uuid.UUID(test_id)
        rows = session.execute(self.__prepared_statements['select_test_artifact_data'], (test_id, artifact_type))
        return namedtuple('Artifact', 'artifact description')(rows[0].artifact.decode("hex"), rows[0].description)

    ################################################################################
    ####  Retrieve tests by status:
    ################################################################################
    def get_test_status_by_cluster(self, status, cluster, direction='ASC', limit=999999999):
        session = self.get_session()
        if cluster is None:
            rows = session.execute(self.__prepared_statements['select_test_status_all'], (status, limit))
        else:
            if direction == 'ASC':
                rows = session.execute(self.__prepared_statements['select_test_status_asc'], (status, cluster, limit))
            elif direction == 'DESC':
                rows = session.execute(self.__prepared_statements['select_test_status_asc'], (status, cluster, limit))
            else:
                raise ValueError('Unknown sort direction: {direction}'.format(**locals()))
        return [self.__test_row_to_dict(r) for r in rows]

    def get_scheduled_tests(self, cluster, limit=999999999):
        return self.get_test_status_by_cluster('scheduled', cluster, 'ASC', limit)

    def get_failed_tests(self, cluster, limit=999999999):
        return self.get_test_status_by_cluster('failed', cluster, 'DESC', limit)

    def get_cancelled_tests(self, cluster, limit=999999999):
        return self.get_test_status_by_cluster('cancelled', cluster, 'DESC', limit)

    def get_next_scheduled_test(self, cluster):
        session = self.get_session()
        rows = session.execute(self.__prepared_statements['select_next_scheduled'], (cluster,))
        tests = [self.__test_row_to_dict(r) for r in rows]
        try:
            return tests[0]
        except IndexError:
            raise NoTestsScheduledError('No tests scheduled for {cluster}'.format(cluster=cluster))

    def get_in_progress_tests(self, cluster, limit=999999999):
        return self.get_test_status_by_cluster('in_progress', cluster, 'ASC', limit)

    def get_completed_tests(self, limit=999999999):
        session = self.get_session()
        rows = session.execute(self.__prepared_statements['select_test_completed'], (limit,))
        return [self.__test_row_to_dict(r) for r in rows]
    
    ################################################################################
    ####  Retrieve tests by user:
    ################################################################################
    def get_test_status_by_user(self, status, user, limit=999999999):
        session = self.get_session()
        rows = session.execute(self.__prepared_statements['select_test_status_by_user'], (status, user, limit))
        return [self.__test_row_to_dict(r) for r in rows]

    def get_user_scheduled_tests(self, user, limit=999999999):
        return self.get_test_status_by_user('scheduled', user, limit)

    def get_user_in_progress_tests(self, user, limit=999999999):
        return self.get_test_status_by_user('in_progress', user, limit)

    def get_user_completed_tests(self, user, limit=999999999):
        return self.get_test_status_by_user('completed', user, limit)

    def get_user_failed_tests(self, user, limit=999999999):
        return self.get_test_status_by_user('failed', user, limit)

    ################################################################################
    #### Cluster Management:
    ################################################################################
    def add_cluster(self, name, num_nodes, description):
        session = self.get_session()
        session.execute(self.__prepared_statements['insert_clusters'], (name, num_nodes, description))

    def get_cluster_names(self):
        session = self.get_session()
        rows = session.execute(self.__prepared_statements['select_clusters_name'], [])
        return [c.name for c in rows]

    ################################################################################
    #### API Keys:
    ################################################################################
    def add_pub_key(self, name, user_type, pubkey):
        """Add an API pubkey with the given name and user_type (cluster, user)"""
        session = self.get_session()
        try:
            existing_key = self.get_pub_key(name)
            raise APIKeyExistsError('API key already exists for {name}'.format(name=name))
        except UnknownAPIKeyError:
            pass
        session.execute(self.__prepared_statements['insert_api_pubkey'], (name, user_type, pubkey))

    def get_pub_key(self, name):
        session = self.get_session()
        rows = session.execute(self.__prepared_statements['select_api_pubkey'], (name,))
        try:
            return rows[0].__dict__
        except IndexError:
            raise UnknownAPIKeyError('No API key named {name}'.format(name=name))

    ################################################################################
    #### User Management:
    ################################################################################
    def create_user(self, user_id, full_name, roles):
        session = self.get_session()
        session.execute(self.__prepared_statements['insert_user'], (user_id, full_name, roles))
        return user_id

    def get_user(self, user_id):
        session = self.get_session()
        try:
            return session.execute(self.__prepared_statements['select_user'], (user_id,))[0]
        except IndexError:
            raise UnknownUserError('Unknown User {user_id}'.format(user_id=user_id))

    def get_user_roles(self, user_id):
        session = self.get_session()
        try:
            return session.execute(self.__prepared_statements['select_user_roles'], (user_id,))[0].roles
        except IndexError:
            raise UnknownUserError('Unknown User {user_id}'.format(user_id=user_id))
    

    def __test_row_to_dict(self, row):
        test = row.__dict__
        #Deserialize test definition:
        if test.has_key('test_definition'):
            test['test_definition'] = json.loads(test['test_definition'])
        #Compute a usable date field from the test_id:
        test['scheduled_date'] = uuid_to_datetime(test['test_id'])
        log.debug(row)
        if test.has_key('completed_date') and test['completed_date'] is not None:
            test['completed_date'] = uuid_to_datetime(test['completed_date'])
        test['test_id'] = str(test['test_id'])
        return test

    def _zmq_show_jobs(self, endpoint=None, subscribe_filter='scheduled bdplab '):
        """Debug utility to show jobs via zeromq

        endpoint - the zeromq endpoint, default to the publisher endpoint of this model object"""
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        if endpoint is None:
            endpoint = self.zmq_endpoint
        socket.connect(endpoint)
        socket.setsockopt_string(zmq.SUBSCRIBE, unicode(subscribe_filter))
        while True:
            print socket.recv_string()



import sys
from setuptools import setup, find_packages

requires = (
    'flask',
    'Flask-Script',
    'flask_sockets',
    'gunicorn',
    'cassandra-driver',
    'google-api-python-client',
    'ecdsa',
    'daemonize',
    'websocket-client',
    'pyzmq',
    'fabric',
    'pyyaml',
    'supervisor',
    'pexpect',
    'blist'
)

setup(
    name = 'cstar_perf.frontend',
    version = '1.0',
    description = 'A web frontend for cstar_perf, the Cassandra performance testing platform',
    author = 'The DataStax Cassandra Test Engineering Team',
    author_email = 'ryan@datastax.com',
    url = 'https://github.com/datastax/cstar_perf',
    install_requires = requires,
    namespace_packages = ['cstar_perf'],
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    entry_points = {'console_scripts': 
                    ['cstar_perf_client = cstar_perf.frontend.client.client:main',
                     'cstar_perf_server = cstar_perf.frontend.lib.server:main',
                     'cstar_perf_notifications = cstar_perf.frontend.server.notifications:main']},
)

from cstar_perf.frontend.lib.crypto import generate_server_keys
generate_server_keys()



from flask import ( Flask, render_template, request, redirect, abort,
                    jsonify, make_response, session)
from app import app, db
from model import UnknownUserError, UnknownTestError

################################################################################
#### Template functions:
################################################################################
def get_user_id():
    return session.get('user_id', None)
app.jinja_env.globals['get_user_id'] = get_user_id

def user_is_authenticated():
    return session.get('logged_in',False)
app.jinja_env.globals['user_is_authenticated'] = user_is_authenticated


################################################################################
#### Helpers
################################################################################
def user_in_role(role, user=None):
    """Find if a user is in the given role"""
    if user is None:
        user = get_user_id()
    try:
        user_roles = db.get_user_roles(get_user_id())
        if role in user_roles:
            return True
    except UnknownUserError:
        pass
    return False

def requires_auth(role):
    """Ensures the current user has the appropriate authorization before
    running the wrapped function"""
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kw):
            # Do the check:
            if user_is_authenticated():
                if user_in_role(role):
                    return function(*args, **kw)
            return make_response(render_template('access_denied.jinja2.html'), 401)
        return wrapper
    return decorator



import datetime
import random
import string
import signal
import time
from contextlib import contextmanager
import os

def uuid_to_datetime(uid):
    return datetime.datetime.fromtimestamp((uid.get_time() - 0x01b21dd213814000L)*100/1e9)

def random_token():
    return ''.join(random.choice(string.ascii_uppercase + string.digits)
                   for x in xrange(32))

class TimeoutError(Exception):
    pass

class timeout:
    """A context manager to timeout a block of code after x seconds

    >>> with timeout(3):
    ...     time.sleep(4)
    Traceback (most recent call last):
        ...
    TimeoutError: Timeout
    """
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message
    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)

def format_bytesize(num):
    """Convert number of bytes to human readable string"""
    for x in ['bytes','KB','MB','GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)
